"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
var react_1 = __importStar(require("react"));
var WarningCode = {
    UnfoundClaim: 1 << 0,
    ContestedClaim: 1 << 1,
    ClaimCorruption: 1 << 2
};
var logNothing = function () { };
var Meta = /** @class */ (function () {
    function Meta() {
        this.claims = new Map();
    }
    Meta.prototype.reclaim = function (before, after, logWarning, path) {
        if (before === void 0) { before = ""; }
        if (after === void 0) { after = ""; }
        if (logWarning === void 0) { logWarning = logNothing; }
        if (path === void 0) { path = ""; }
        var claims = this.claims;
        var prior = claims.get(before) || 0;
        var next = claims.get(after) || 0;
        var warning = 0;
        if (before !== after) {
            if (prior > 0) {
                if (prior === 1) {
                    claims.delete(before);
                }
                else {
                    claims.set(before, prior - 1);
                }
            }
            else if (before !== "") {
                warning |= WarningCode.UnfoundClaim;
                logWarning("Cannot restore unfound claim: " + before, path);
            }
            if (after !== "") {
                claims.set(after, next + 1);
                if (next > 0) {
                    warning |= WarningCode.ContestedClaim;
                    logWarning("Contested claim: " + after, path);
                }
                else if (next < 0) {
                    warning |= WarningCode.ClaimCorruption;
                    logWarning("Corrupt claim" + after, path);
                }
            }
        }
        return warning;
    };
    return Meta;
}());
var useFinalMeta = function () {
    var ref = react_1.useRef(null);
    ref.current = ref.current || new Meta();
    return ref.current;
};
var idCounter = 0;
var useFinalCommonId = function () {
    var ref = react_1.useRef(null);
    ref.current = ref.current || --idCounter;
    return ref.current;
};
var logRed = function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    console.error.apply(console, args);
};
exports.createSystem = function (_a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.separator, separator = _c === void 0 ? "__" : _c, _d = _b.useFinalId, useFinalId = _d === void 0 ? useFinalCommonId : _d, _e = _b.logWarning, logWarning = _e === void 0 ? logRed : _e;
    var Directory = react_1.createContext({
        name: "",
        id: null,
        path: separator,
        meta: new Meta()
    });
    var Folder = function (props) {
        var _a = props.name, name = _a === void 0 ? "" : _a, children = props.children;
        var id = useFinalId();
        var meta = useFinalMeta();
        var subdir = useCWDKey(name) + (name === "" ? "" : name + separator);
        var ref = react_1.useRef(null);
        var _b = ref.current, cached = _b === void 0 ? null : _b;
        // TODO: consider signalling changes through callback props
        if (typeof name !== "string") {
            throw new Error("Folder name must be a string");
        }
        else if (name.indexOf(separator) > -1) {
            throw new Error("Folder name forbids separator: " + separator + " ( " + name + " )");
        }
        if (!cached || cached.path !== subdir || cached.name !== name) {
            cached = {
                name: name,
                id: id,
                path: subdir,
                meta: meta
            };
        }
        ref.current = cached;
        return (react_1.default.createElement(Directory.Provider, { value: ref.current }, typeof children === "function" ? children({ path: subdir }) : children));
    };
    var useCWDKey = function (key) {
        if (key === void 0) { key = ""; }
        var parent = react_1.useContext(Directory);
        react_1.useEffect(function () {
            parent.meta.reclaim("", key, logWarning, parent.path);
            return function () {
                parent.meta.reclaim(key, "", logWarning, parent.path);
            };
        }, [key, parent.meta]); // NOTE: parent.path is excluded
        return parent.path;
    };
    var useCWDRef = function () {
        var dir = react_1.useContext(Directory);
        var ref = react_1.useRef(null);
        ref.current = ref.current || {
            path: dir ? dir.path : null,
            working: function () { return ref.current.path; }
        };
        ref.current.path = dir ? dir.path : null;
        return ref.current.working;
    };
    var mkDir = function (Component) { return function (props) {
        var _a = props.folder, folder = _a === void 0 ? "" : _a;
        return folder === "" ? (react_1.default.createElement(Component, __assign({}, props))) : (react_1.default.createElement(Folder, { name: folder },
            react_1.default.createElement(Component, __assign({}, props))));
    }; };
    // TODO: consider adding state management with Files
    // const useFiles = () => {}
    return [Folder, mkDir, useCWDRef];
};
exports.Folder = (_a = exports.createSystem(), _a[0]), exports.mkDir = _a[1], exports.useCWDRef = _a[2];
